@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -141,17 +141,16 @@ void run_instance
 		delete ring;
 		exit(-1);
 	}
+	delete ring;
 	if (!prv->pub->valid || prv->Weak(opt_verbose))
 	{
 		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	if (prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7)
 	{
 		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -161,7 +160,6 @@ void run_instance
 	if (!init_tDSS(prv, opt_verbose, dss))
 	{
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -170,7 +168,6 @@ void run_instance
 	{
 		std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" << std::endl;
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -186,10 +183,9 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read" <<
-					" password for protecting channel to R_" << i << std::endl;
+				std::cerr << "ERROR: p_" << whoami << ": cannot read" <<
+					" password for protecting channel to p_" << i << std::endl;
 				delete dss;
-				delete ring;
 				delete prv;
 				exit(-1);
 			}
@@ -197,11 +193,10 @@ void run_instance
 			if (((i + 1) < peers.size()) &&
 				!TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip" <<
-					" to next password for protecting channel to R_" <<
+				std::cerr << "ERROR: p_" << whoami << ": cannot skip" <<
+					" to next password for protecting channel to p_" <<
 					(i + 1) << std::endl;
 				delete dss;
-				delete ring;
 				delete prv;
 				exit(-1);
 			}
@@ -209,7 +204,7 @@ void run_instance
 		else
 		{
 			// use simple key -- we assume that GNUnet provides secure channels
-			key << "dkg-refresh::R_" << (i + whoami);
+			key << "dkg-refresh::p_" << (i + whoami);
 		}
 		uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
@@ -243,76 +238,83 @@ void run_instance
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
-	for (size_t i = 0; i < num_xtests; i++)
-	{
-		mpz_t xtest;
-		mpz_init_set_ui(xtest, i);
-		std::cerr << "INFO: R_" << whoami << ": xtest = " << xtest << " <-> ";
-		rbc->Broadcast(xtest);
-		for (size_t ii = 0; ii < peers.size(); ii++)
-		{
-			if (!rbc->DeliverFrom(xtest, ii))
-				std::cerr << "<X> ";
-			else
-				std::cerr << xtest << " ";
-		}
-		std::cerr << std::endl;
-		mpz_clear(xtest);
-	}
+	xtest(num_xtests, whoami, peers.size(), rbc);
 
 	// update the tDSS key (proactive security against mobile adversary)
 	std::stringstream err_log;
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": dss.Refresh()" << std::endl;
+		std::cerr << "INFO: p_" << whoami << ": dss.Refresh()" << std::endl;
 	if (!dss->Refresh(peers.size(), whoami, prv->tdss_idx2dkg,
 		prv->tdss_dkg2idx, aiou, rbc, err_log, false, cache, cache_mod,
 		&cache_avail))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Refresh() failed" <<
+		std::cerr << "ERROR: p_" << whoami << ": " << "tDSS Refresh() failed" <<
 			std::endl;
-		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl <<
+		std::cerr << "ERROR: p_" << whoami << ": log follows " << std::endl <<
 			err_log.str();
 		delete rbc, delete aiou, delete aiou2;
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log.str();
+		std::cerr << "INFO: p_" << whoami << ": log follows " <<
+			std::endl << err_log.str();
+
 	// at the end: deliver some more rounds for still waiting parties
-	time_t synctime = aiounicast::aio_timeout_long;
+	time_t synctime = (opt_W * 6);
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": waiting approximately " <<
+		std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
 			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
 			std::endl;
 	rbc->Sync(synctime);
-	// create an OpenPGP DSA-based primary key using refreshed values from tDSS
+
+	// release RBC
+	delete rbc;
+	
+	// release handles (unicast channel)
+	uP_in.clear(), uP_out.clear(), uP_key.clear();
+	if (opt_verbose)
+	{
+		std::cerr << "INFO: p_" << whoami << ": unicast channels";
+		aiou->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
+
+	// release handles (broadcast channel)
+	bP_in.clear(), bP_out.clear(), bP_key.clear();
+	if (opt_verbose)
+	{
+		std::cerr << "INFO: p_" << whoami << ": broadcast channel";
+		aiou2->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
+
+	// release asynchronous unicast and broadcast
+	delete aiou, delete aiou2;
+
+	// create an OpenPGP secret key packet using refreshed values from tDSS
 	gcry_mpi_t n, t, i, qualsize, x_rvss_qualsize, x_i, xprime_i;
 	std::vector<gcry_mpi_t> qual, x_rvss_qual;
 	std::vector< std::vector<gcry_mpi_t> > c_ik;
 	x_i = gcry_mpi_snew(2048);
 	if (!tmcg_mpz_get_gcry_mpi(x_i, dss->x_i))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+		std::cerr << "ERROR: p_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
 			" failed for x_i" << std::endl;
 		gcry_mpi_release(x_i);
-		delete rbc, delete aiou, delete aiou2;
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	xprime_i = gcry_mpi_snew(2048);
 	if (!tmcg_mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+		std::cerr << "ERROR: p_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
 			" failed for xprime_i" << std::endl;
 		gcry_mpi_release(x_i);
 		gcry_mpi_release(xprime_i);
-		delete rbc, delete aiou, delete aiou2;
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -340,7 +342,7 @@ void run_instance
 			tmp = gcry_mpi_new(2048);
 			if (!tmcg_mpz_get_gcry_mpi(tmp, dss->dkg->x_rvss->C_ik[j][k]))
 			{
-				std::cerr << "ERROR: R_" << whoami <<
+				std::cerr << "ERROR: p_" << whoami <<
 					": tmcg_mpz_get_gcry_mpi() failed for" <<
 					" dss->dkg->x_rvss->C_ik[j][k]" << std::endl;
 				gcry_mpi_release(x_i);
@@ -358,21 +360,19 @@ void run_instance
 					for (size_t kk = 0; kk < c_ik[jj].size(); kk++)
 						gcry_mpi_release(c_ik[jj][kk]);
 				gcry_mpi_release(tmp);
-				delete rbc, delete aiou, delete aiou2;
 				delete dss;
-				delete ring;
 				delete prv;
 				exit(-1); 
 			}
 			c_ik[j].push_back(tmp);
 		}
 	}
-	tmcg_openpgp_octets_t sec;
+	tmcg_openpgp_octets_t pkt;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		PacketSecEncodeExperimental107(prv->pub->creationtime, prv->pub->dsa_p,
 			prv->pub->dsa_q, prv->pub->dsa_g, prv->tdss_h, prv->pub->dsa_y, n,
 			t, i, qualsize, qual, x_rvss_qualsize, x_rvss_qual, prv->tdss_capl,
-			c_ik, x_i, xprime_i, passphrase, sec);
+			c_ik, x_i, xprime_i, passphrase, pkt);
 	gcry_mpi_release(x_i);
 	gcry_mpi_release(xprime_i);
 	gcry_mpi_release(n);
@@ -387,131 +387,27 @@ void run_instance
 	for (size_t j = 0; j < c_ik.size(); j++)
 		for (size_t k = 0; k < c_ik[j].size(); k++)
 			gcry_mpi_release(c_ik[j][k]);
-	
-	// export updated private key in OpenPGP armor format
-	tmcg_openpgp_octets_t all;
-	std::string armor;
-	std::stringstream secfilename;
-	secfilename << peers[whoami] << "_dkg-sec.asc";
-	all.insert(all.end(), sec.begin(), sec.end());
-	for (size_t k = 0; k < prv->pub->selfsigs.size(); k++)
-		all.insert(all.end(),
-			(prv->pub->selfsigs[k]->packet).begin(),
-			(prv->pub->selfsigs[k]->packet).end());
-	for (size_t k = 0; k < prv->pub->keyrevsigs.size(); k++)
-		all.insert(all.end(),
-			(prv->pub->keyrevsigs[k]->packet).begin(),
-			(prv->pub->keyrevsigs[k]->packet).end());
-	for (size_t j = 0; j < prv->pub->userids.size(); j++)
-	{
-		TMCG_OpenPGP_UserID *uid = prv->pub->userids[j];
-		if (uid->valid)
-		{
-			all.insert(all.end(),
-				(uid->packet).begin(), (uid->packet).end());
-			for (size_t k = 0; k < uid->selfsigs.size(); k++)
-				all.insert(all.end(),
-					(uid->selfsigs[k]->packet).begin(),
-					(uid->selfsigs[k]->packet).end());
-			for (size_t k = 0; k < uid->revsigs.size(); k++)
-				all.insert(all.end(),
-					(uid->selfsigs[k]->packet).begin(),
-					(uid->selfsigs[k]->packet).end());
-		}
-	}
-	for (size_t j = 0; j < prv->pub->userattributes.size(); j++)
-	{
-		TMCG_OpenPGP_UserAttribute *uat = prv->pub->userattributes[j];
-		if (uat->valid)
-		{
-			all.insert(all.end(),
-				(uat->packet).begin(), (uat->packet).end());
-			for (size_t k = 0; k < uat->selfsigs.size(); k++)
-				all.insert(all.end(),
-					(uat->selfsigs[k]->packet).begin(),
-					(uat->selfsigs[k]->packet).end());
-			for (size_t k = 0; k < uat->revsigs.size(); k++)
-				all.insert(all.end(),
-					(uat->selfsigs[k]->packet).begin(),
-					(uat->selfsigs[k]->packet).end());
-		}
-	}
-	for (size_t j = 0; j < prv->private_subkeys.size(); j++)
-	{
-		TMCG_OpenPGP_PrivateSubkey *ssb = prv->private_subkeys[j];
-		all.insert(all.end(), (ssb->packet).begin(), (ssb->packet).end());
-		TMCG_OpenPGP_Subkey *sub = prv->private_subkeys[j]->pub;
-		for (size_t k = 0; k < sub->selfsigs.size(); k++)
-			all.insert(all.end(),
-				(sub->selfsigs[k]->packet).begin(),
-				(sub->selfsigs[k]->packet).end());
-		for (size_t k = 0; k < sub->bindsigs.size(); k++)
-			all.insert(all.end(),
-				(sub->bindsigs[k]->packet).begin(),
-				(sub->bindsigs[k]->packet).end());
-		for (size_t k = 0; k < sub->pbindsigs.size(); k++)
-			all.insert(all.end(),
-				(sub->pbindsigs[k]->packet).begin(),
-				(sub->pbindsigs[k]->packet).end());
-		for (size_t k = 0; k < sub->keyrevsigs.size(); k++)
-			all.insert(all.end(),
-				(sub->keyrevsigs[k]->packet).begin(),
-				(sub->keyrevsigs[k]->packet).end());
-	}
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
-	if (opt_verbose > 1)
-		std::cout << armor << std::endl;
-	std::ofstream secofs((secfilename.str()).c_str(),
-		std::ofstream::out | std::ofstream::trunc);
-	if (!secofs.good())
-	{
-		std::cerr << "ERROR: R_" << whoami << ": opening private key file failed" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	secofs << armor;
-	if (!secofs.good())
-	{
-		std::cerr << "ERROR: R_" << whoami << ": writing private key file failed" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	secofs.close();
-
-	// release RBC
-	delete rbc;
-	
-	// release handles (unicast channel)
-	uP_in.clear(), uP_out.clear(), uP_key.clear();
-	if (opt_verbose)
-	{
-		std::cerr << "INFO: R_" << whoami << ": unicast channels";
-		aiou->PrintStatistics(std::cerr);
-		std::cerr << std::endl;
-	}
-
-	// release handles (broadcast channel)
-	bP_in.clear(), bP_out.clear(), bP_key.clear();
-	if (opt_verbose)
-	{
-		std::cerr << "INFO: R_" << whoami << ": broadcast channel";
-		aiou2->PrintStatistics(std::cerr);
-		std::cerr << std::endl;
-	}
-
-	// release asynchronous unicast and broadcast
-	delete aiou, delete aiou2;
 
 	// release
 	delete dss;
-	delete ring;
+	
+	// replace secret key packet in the existing OpenPGP structure of this key
+	prv->packet.clear();
+	prv->packet.insert(prv->packet.end(), pkt.begin(), pkt.end());
+
+	// export and write updated private key in OpenPGP armor format
+	std::stringstream secfilename;
+	secfilename << peers[whoami] << "_dkg-sec.asc";
+	tmcg_openpgp_octets_t sec;
+	prv->Export(sec);
+	if (!write_key_file(secfilename.str(),
+		TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, sec))
+	{
+		delete prv;
+		exit(-1);
+	}
+
+	// release
 	delete prv;
 }
 
@@ -535,16 +431,16 @@ void fork_instance
 	{
 		if (pid[whoami] == 0)
 		{
-			/* BEGIN child code: participant R_i */
+			/* BEGIN child code: participant p_i */
 #ifdef GNUNET
 			run_instance(whoami, gnunet_opt_xtests);
 #else
 			run_instance(whoami, 0);
 #endif
 			if (opt_verbose)
-				std::cerr << "INFO: R_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: p_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
-			/* END child code: participant R_i */
+			/* END child code: participant p_i */
 		}
 		else
 		{
@@ -600,13 +496,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of refresh protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -655,7 +551,8 @@ int main
 			(arg.find("-k") == 0))
 		{
 			size_t idx = ++i;
-			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) && (opt_hostname == NULL))
+			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_hostname == NULL))
 			{
 				hostname = argv[i+1];
 				opt_hostname = (char*)hostname.c_str();
@@ -666,41 +563,57 @@ int main
 				kfilename = argv[i+1];
 				opt_k = (char*)kfilename.c_str();
 			}
-			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) && (opt_passwords == NULL))
+			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_passwords == NULL))
 			{
 				passwords = argv[i+1];
 				opt_passwords = (char*)passwords.c_str();
 			}
-			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) && (port.length() == 0))
+			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) &&
+				(port.length() == 0))
+			{
 				port = argv[i+1];
-			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) && (opt_W == 5))
+			}
+			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_W == 5))
+			{
 				opt_W = strtoul(argv[i+1], NULL, 10);
+			}
 			continue;
 		}
-		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-V") == 0))
+		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) ||
+			(arg.find("-h") == 0) || (arg.find("-V") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
 #ifndef GNUNET
 				std::cout << usage << std::endl;
 				std::cout << about << std::endl;
-				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
+				std::cout << "Arguments mandatory for long options are also" <<
+					" mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
-				std::cout << "  -H STRING      hostname (e.g. onion address) of this peer within PEERS" << std::endl;
+				std::cout << "  -H STRING      hostname (e.g. onion address)" <<
+					" of this peer within PEERS" << std::endl;
 				std::cout << "  -k FILENAME    use keyring FILENAME" <<
 					" containing external revocation keys" << std::endl;
-				std::cout << "  -p INTEGER     start port for built-in TCP/IP message exchange service" << std::endl;
-				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
-				std::cout << "  -v, --version  print the version number" << std::endl;
-				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point messages in minutes" << std::endl;
+				std::cout << "  -p INTEGER     start port for built-in" <<
+					" TCP/IP message exchange service" << std::endl;
+				std::cout << "  -P STRING      exchanged passwords to" <<
+					" protect private and broadcast channels" << std::endl;
+				std::cout << "  -v, --version  print the version number" <<
+					std::endl;
+				std::cout << "  -V, --verbose  turn on verbose output" <<
+					std::endl;
+				std::cout << "  -W INTEGER     timeout for point-to-point" <<
+					" messages in minutes" << std::endl;
 #endif
 				return 0; // not continue
 			}
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
 			{
 #ifndef GNUNET
-				std::cout << "dkg-refresh v" << version << " without GNUNET support" << std::endl;
+				std::cout << "dkg-refresh v" << version <<
+					" without GNUNET support" << std::endl;
 #endif
 				return 0; // not continue
 			}
@@ -720,7 +633,8 @@ int main
 		}
 		else
 		{
-			std::cerr << "ERROR: peer identity \"" << arg << "\" too long" << std::endl;
+			std::cerr << "ERROR: peer identity \"" << arg << "\" too long" <<
+				std::endl;
 			return -1;
 		}
 	}
@@ -734,19 +648,17 @@ int main
 	// check command line arguments
 	if ((opt_hostname != NULL) && (opt_passwords == NULL))
 	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		std::cerr << "ERROR: option \"-P\" is necessary due to insecure" <<
+			" network" << std::endl;
 		return -1;
 	}
 	if (peers.size() < 1)
 	{
-		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
+		std::cerr << "ERROR: no peers given as argument; usage: " <<
+			usage << std::endl;
 		return -1;
 	}
-
-	// canonicalize peer list
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
+	canonicalize(peers);
 	if ((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N))
 	{
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
@@ -760,7 +672,7 @@ int main
 	}
 
 	// lock memory
-	bool force_secmem = false;
+	bool force_secmem = false, should_unlock = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
@@ -768,15 +680,22 @@ int main
 		// at least try to use libgcrypt's secure memory
 		force_secmem = true;
 	}
+	else
+		should_unlock = true;
 
 	// initialize LibTMCG
 	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	if (opt_verbose)
-		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+	{
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
+			std::endl;
+	}
 
 	// read the public key
 	std::string armored_pubkey = "undefined";
@@ -795,7 +714,10 @@ int main
 	}
 	if (armored_pubkey == "undefined")
 	{
-		std::cerr << "ERROR: no corresponding public key file found" << std::endl;
+		std::cerr << "ERROR: no corresponding public key file found" <<
+			std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 
@@ -809,12 +731,16 @@ int main
 		{
 			std::cerr << "ERROR: primary key is not DSA" << std::endl;
 			delete primary;
+			if (should_unlock)
+				unlock_memory();
 			return -1;
 		}
 	}
 	else
 	{
 		std::cerr << "ERROR: cannot use the provided public key" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 
@@ -827,165 +753,95 @@ int main
 			std::endl;
 		mpz_clear(dss_q);
 		delete primary;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	delete primary;
-	std::cerr << "We need some entropy to cache very strong randomness for share refresh." << std::endl;
-	std::cerr << "Please use other programs, move the mouse, and type on your keyboard: " << std::endl; 
-	tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail, (2 * peers.size()), dss_q);
+	std::cerr << "We need some entropy to cache very strong randomness for" <<
+		" share refresh." << std::endl;
+	std::cerr << "Please use other programs, move the mouse, and type on" <<
+		" your keyboard: " << std::endl; 
+	tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, cache_avail,
+		(2 * peers.size()), dss_q);
 	std::cerr << "Thank you!" << std::endl;
 	mpz_clear(dss_q);
 
-	// initialize return code
+	// initialize return code and do the main work
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP service requested
 	if (opt_hostname != NULL)
 	{
-		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // get start port from options
-		if ((opt_p < 1) || (opt_p > 65535))
-		{
-			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
-			return -1;
-		}
-		tcpip_init(hostname);
-		tcpip_bindports((uint16_t)opt_p, false);
-		tcpip_bindports((uint16_t)opt_p, true);
-		while (tcpip_connect((uint16_t)opt_p, false) < peers.size())
-			sleep(1);
-		while (tcpip_connect((uint16_t)opt_p, true) < peers.size())
-			sleep(1);
-		tcpip_accept();
-		tcpip_fork();
-		ret = tcpip_io();
-		tcpip_close();
-		tcpip_done();
-		// release cache
-		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
-		return ret;
+		// start interactive variant, if built-in TCP/IP requested
+		ret = run_tcpip(peers.size(), opt_p, hostname, port);
 	}
-
-	// start interactive variant with GNUnet or otherwise a local test
-#ifdef GNUNET
-	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
-		GNUNET_GETOPT_option_string('H',
-			"hostname",
-			"STRING",
-			"hostname (e.g. onion address) of this peer within PEERS",
-			&gnunet_opt_hostname
-		),
-		GNUNET_GETOPT_option_string('k',
-			"keyring",
-			"FILENAME",
-			"use keyring FILENAME containing external revocation keys",
-			&gnunet_opt_k
-		),
-		GNUNET_GETOPT_option_string('p',
-			"port",
-			"STRING",
-			"GNUnet CADET port to listen/connect",
-			&gnunet_opt_port
-		),
-		GNUNET_GETOPT_option_string('P',
-			"passwords",
-			"STRING",
-			"exchanged passwords to protect private and broadcast channels",
-			&gnunet_opt_passwords
-		),
-		GNUNET_GETOPT_option_flag('V',
-			"verbose",
-			"turn on verbose output",
-			&gnunet_opt_verbose
-		),
-		GNUNET_GETOPT_option_uint('w',
-			"wait",
-			"TIME",
-			"minutes to wait until start of refresh protocol",
-			&gnunet_opt_wait
-		),
-		GNUNET_GETOPT_option_uint('W',
-			"aiou-timeout",
-			"TIME",
-			"timeout for point-to-point messages in minutes",
-			&gnunet_opt_W
-		),
-		GNUNET_GETOPT_option_uint('x',
-			"x-tests",
-			NULL,
-			"number of exchange tests",
-			&gnunet_opt_xtests
-		),
-		GNUNET_GETOPT_OPTION_END
-	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run, argv[0]);
-//	GNUNET_free((void *) argv);
-	// release cache
-	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
-	if (ret == GNUNET_OK)
-		return 0;
 	else
-		return -1;
+	{
+		// start interactive variant with GNUnet or otherwise a local test
+#ifdef GNUNET
+		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_string('H',
+				"hostname",
+				"STRING",
+				"hostname (e.g. onion address) of this peer within PEERS",
+				&gnunet_opt_hostname
+			),
+			GNUNET_GETOPT_option_string('k',
+				"keyring",
+				"FILENAME",
+				"use keyring FILENAME containing external revocation keys",
+				&gnunet_opt_k
+			),
+			GNUNET_GETOPT_option_string('p',
+				"port",
+				"STRING",
+				"GNUnet CADET port to listen/connect",
+				&gnunet_opt_port
+			),
+			GNUNET_GETOPT_option_string('P',
+				"passwords",
+				"STRING",
+				"exchanged passwords to protect private and broadcast channels",
+				&gnunet_opt_passwords
+			),
+			GNUNET_GETOPT_option_flag('V',
+				"verbose",
+				"turn on verbose output",
+				&gnunet_opt_verbose
+			),
+			GNUNET_GETOPT_option_uint('w',
+				"wait",
+				"INTEGER",
+				"minutes to wait until start of refresh protocol",
+				&gnunet_opt_wait
+			),
+			GNUNET_GETOPT_option_uint('W',
+				"aiou-timeout",
+				"INTEGER",
+				"timeout for point-to-point messages in minutes",
+				&gnunet_opt_W
+			),
+			GNUNET_GETOPT_option_uint('x',
+				"x-tests",
+				NULL,
+				"number of exchange tests",
+				&gnunet_opt_xtests
+			),
+			GNUNET_GETOPT_OPTION_END
+		};
+		ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions,
+			&gnunet_run, argv[0]);
+		if (ret != GNUNET_OK)
+			ret = -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
+			broadcast_pipefd, &fork_instance);
 #endif
-
-	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
-	std::cerr << "WARNING: due to cache issues the generated shares are identical, don't use them!" << std::endl;
-	// open pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if (pipe(pipefd[i][j]) < 0)
-				perror("ERROR: dkg-refresh (pipe)");
-			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("ERROR: dkg-refresh (pipe)");
-		}
 	}
-	
-	// start childs
-	for (size_t i = 0; i < peers.size(); i++)
-		fork_instance(i);
-
-	// sleep for five seconds
-	sleep(5);
-	
-	// wait for childs and close pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		int wstatus = 0;
-		if (opt_verbose)
-			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
-		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("ERROR: dkg-refresh (waitpid)");
-		if (!WIFEXITED(wstatus))
-		{
-			std::cerr << "ERROR: protocol instance ";
-			if (WIFSIGNALED(wstatus))
-				std::cerr << pid[i] << " terminated by signal " << WTERMSIG(wstatus) << std::endl;
-			if (WCOREDUMP(wstatus))
-				std::cerr << pid[i] << " dumped core" << std::endl;
-			ret = -1; // fatal error
-		}
-		else if (WIFEXITED(wstatus))
-		{
-			if (opt_verbose)
-				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
-			if (WEXITSTATUS(wstatus))
-				ret = -2; // error
-		}
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-refresh (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-refresh (close)");
-		}
-	}
-	
 	// release cache
-	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
-	// finish	
+	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
+	// finish
+	if (should_unlock)
+		unlock_memory();
 	return ret;
 }
 
