@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -114,8 +114,11 @@ void run_instance
 	bool parse_ok;
 	if (opt_k)
 	{
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
-			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+			PublicKeyringParse(armored_pubring, opt_verbose_ring, ring);
 		if (!parse_ok)
 		{
 			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
@@ -262,13 +265,7 @@ void run_instance
 		// participants must agree on a common signature creation time (OpenPGP)
 		csigtime = agree_time(sigtime, whoami, peers.size(), opt_verbose, rbc);
 		// select hash algorithm for OpenPGP based on |q| (size in bit)
-		if (mpz_sizeinbase(dss->q, 2L) == 256)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-		else if (mpz_sizeinbase(dss->q, 2L) == 384)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-		else if (mpz_sizeinbase(dss->q, 2L) == 512)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
-		else
+		if (!select_hashalgo(dss, hashalgo))
 		{
 			std::cerr << "ERROR: p_" << whoami << ": selecting hash" <<
 				" algorithm failed for |q| = " << mpz_sizeinbase(dss->q, 2L) <<
@@ -289,9 +286,7 @@ void run_instance
 	if (opt_verbose)
 		std::cerr << "INFO: hashing the input file \"" << opt_ifilename <<
 			"\"" << std::endl;
-	tmcg_openpgp_octets_t trailer, hash, left, issuer;
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(prv->pub->pub_hashing, issuer);	
+	tmcg_openpgp_octets_t trailer, hash, left;
 	if (opt_t || opt_C)
 	{
 		if (opt_y == NULL)
@@ -299,14 +294,16 @@ void run_instance
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareDetachedSignature(
 					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		}
 		else
 		{
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareDetachedSignature(
 					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		}
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			TextDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
@@ -330,14 +327,16 @@ void run_instance
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareDetachedSignature(
 					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		}
 		else
 		{
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareDetachedSignature(
 					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		}
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
@@ -373,7 +372,7 @@ void run_instance
 	if (opt_y == NULL)
 	{
 		// at the end: deliver some more rounds for still waiting parties
-		time_t synctime = aiounicast::aio_timeout_long;
+		time_t synctime = (opt_W * 6);
 		if (opt_verbose)
 			std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
 				(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
@@ -825,12 +824,7 @@ int main
 			usage << std::endl;
 		return -1;
 	}
-
-	// canonicalize peer list
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it =
-		std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
+	canonicalize(peers);
 	if (((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N)) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
@@ -844,7 +838,7 @@ int main
 	}
 
 	// lock memory
-	bool force_secmem = false;
+	bool force_secmem = false, should_unlock = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
@@ -852,223 +846,145 @@ int main
 		// at least try to use libgcrypt's secure memory
 		force_secmem = true;
 	}
+	else
+		should_unlock = true;
 
 	// initialize LibTMCG
 	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	if (opt_verbose)
 		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
 			std::endl;
 	
-	// initialize return code
+	// initialize return code and do the main work
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if ((opt_hostname != NULL) && (opt_y == NULL))
 	{
-		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
-		if ((opt_p < 1) || (opt_p > 65535))
-		{
-			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
-			return -1;
-		}
-		tcpip_init(hostname);
-		tcpip_bindports((uint16_t)opt_p, false);
-		tcpip_bindports((uint16_t)opt_p, true);
-		while (tcpip_connect((uint16_t)opt_p, false) < peers.size())
-			sleep(1);
-		while (tcpip_connect((uint16_t)opt_p, true) < peers.size())
-			sleep(1);
-		tcpip_accept();
-		tcpip_fork();
-		ret = tcpip_io();
-		tcpip_close();
-		tcpip_done();
-		return ret;
+		// start interactive variant, if built-in TCP/IP requested
+		ret = run_tcpip(peers.size(), opt_p, hostname, port);
 	}
 	else if (opt_y != NULL)
 	{
-		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
-		run_instance(0, time(NULL), opt_e, 0);
-		return ret;
+		// start a single instance as replacement for GnuPG et al.
+		fork_instance(0);
+		ret = wait_instance(0, opt_verbose, pid);
 	}
-
-	// start interactive variant with GNUnet or otherwise a local test
-#ifdef GNUNET
-	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
-		GNUNET_GETOPT_option_flag('C',
-			"clear",
-			"apply cleartext signature framework (cf. RFC 4880)",
-			&gnunet_opt_C
-		),
-		GNUNET_GETOPT_option_uint('e',
-			"expiration",
-			"INTEGER",
-			"expiration time of generated signature in seconds",
-			&gnunet_opt_sigexptime
-		),
-		GNUNET_GETOPT_option_flag('E',
-			"echo",
-			"enable terminal echo when reading passphrase",
-			&gnunet_opt_E
-		),
-		GNUNET_GETOPT_option_string('H',
-			"hostname",
-			"STRING",
-			"hostname (e.g. onion address) of this peer within PEERS",
-			&gnunet_opt_hostname
-		),
-		GNUNET_GETOPT_option_string('i',
-			"input",
-			"FILENAME",
-			"create signature from FILENAME",
-			&gnunet_opt_ifilename
-		),
-		GNUNET_GETOPT_option_string('k',
-			"keyring",
-			"FILENAME",
-			"use keyring FILENAME containing external revocation keys",
-			&gnunet_opt_k
-		),
-		GNUNET_GETOPT_option_string('o',
-			"output",
-			"FILENAME",
-			"write generated signature to FILENAME",
-			&gnunet_opt_ofilename
-		),
-		GNUNET_GETOPT_option_string('p',
-			"port",
-			"STRING",
-			"GNUnet CADET port to listen/connect",
-			&gnunet_opt_port
-		),
-		GNUNET_GETOPT_option_string('P',
-			"passwords",
-			"STRING",
-			"exchanged passwords to protect private and broadcast channels",
-			&gnunet_opt_passwords
-		),
-		GNUNET_GETOPT_option_string('U',
-			"URI",
-			"STRING",
-			"policy URI tied to signature",
-			&gnunet_opt_URI
-		),
-		GNUNET_GETOPT_option_flag('t',
-			"text",
-			"create canonical text document signature",
-			&gnunet_opt_t
-		),
-		GNUNET_GETOPT_option_flag('V',
-			"verbose",
-			"turn on verbose output",
-			&gnunet_opt_verbose
-		),
-		GNUNET_GETOPT_option_uint('w',
-			"wait",
-			"INTEGER",
-			"minutes to wait until start of signing protocol",
-			&gnunet_opt_wait
-		),
-		GNUNET_GETOPT_option_uint('W',
-			"aiou-timeout",
-			"INTEGER",
-			"timeout for point-to-point messages in minutes",
-			&gnunet_opt_W
-		),
-		GNUNET_GETOPT_option_string('y',
-			"yaot",
-			"FILNAME",
-			"yet another OpenPGP tool with private key in FILENAME",
-			&gnunet_opt_y
-		),
-		GNUNET_GETOPT_option_uint('x',
-			"x-tests",
-			NULL,
-			"number of exchange tests",
-			&gnunet_opt_xtests
-		),
-		GNUNET_GETOPT_OPTION_END
-	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
-		argv[0]);
-//	GNUNET_free((void *) argv);
-	if (ret == GNUNET_OK)
-		return 0;
 	else
-		return -1;
+	{
+		// start interactive variant with GNUnet or otherwise a local test
+#ifdef GNUNET
+		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_flag('C',
+				"clear",
+				"apply cleartext signature framework (cf. RFC 4880)",
+				&gnunet_opt_C
+			),
+			GNUNET_GETOPT_option_uint('e',
+				"expiration",
+				"INTEGER",
+				"expiration time of generated signature in seconds",
+				&gnunet_opt_sigexptime
+			),
+			GNUNET_GETOPT_option_flag('E',
+				"echo",
+				"enable terminal echo when reading passphrase",
+				&gnunet_opt_E
+			),
+			GNUNET_GETOPT_option_string('H',
+				"hostname",
+				"STRING",
+				"hostname (e.g. onion address) of this peer within PEERS",
+				&gnunet_opt_hostname
+			),
+			GNUNET_GETOPT_option_string('i',
+				"input",
+				"FILENAME",
+				"create signature from FILENAME",
+				&gnunet_opt_ifilename
+			),
+			GNUNET_GETOPT_option_string('k',
+				"keyring",
+				"FILENAME",
+				"use keyring FILENAME containing external revocation keys",
+				&gnunet_opt_k
+			),
+			GNUNET_GETOPT_option_string('o',
+				"output",
+				"FILENAME",
+				"write generated signature to FILENAME",
+				&gnunet_opt_ofilename
+			),
+			GNUNET_GETOPT_option_string('p',
+				"port",
+				"STRING",
+				"GNUnet CADET port to listen/connect",
+				&gnunet_opt_port
+			),
+			GNUNET_GETOPT_option_string('P',
+				"passwords",
+				"STRING",
+				"exchanged passwords to protect private and broadcast channels",
+				&gnunet_opt_passwords
+			),
+			GNUNET_GETOPT_option_string('U',
+				"URI",
+				"STRING",
+				"policy URI tied to signature",
+				&gnunet_opt_URI
+			),
+			GNUNET_GETOPT_option_flag('t',
+				"text",
+				"create canonical text document signature",
+				&gnunet_opt_t
+			),
+			GNUNET_GETOPT_option_flag('V',
+				"verbose",
+				"turn on verbose output",
+				&gnunet_opt_verbose
+			),
+			GNUNET_GETOPT_option_uint('w',
+				"wait",
+				"INTEGER",
+				"minutes to wait until start of signing protocol",
+				&gnunet_opt_wait
+			),
+			GNUNET_GETOPT_option_uint('W',
+				"aiou-timeout",
+				"INTEGER",
+				"timeout for point-to-point messages in minutes",
+				&gnunet_opt_W
+			),
+			GNUNET_GETOPT_option_string('y',
+				"yaot",
+				"FILNAME",
+				"yet another OpenPGP tool with private key in FILENAME",
+				&gnunet_opt_y
+			),
+			GNUNET_GETOPT_option_uint('x',
+				"x-tests",
+				NULL,
+				"number of exchange tests",
+				&gnunet_opt_xtests
+			),
+			GNUNET_GETOPT_OPTION_END
+		};
+		ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions,
+			&gnunet_run, argv[0]);
+		if (ret != GNUNET_OK)
+			ret = -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
-		" of this program" << std::endl;
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
+			broadcast_pipefd, &fork_instance);
 #endif
-
-	std::cerr << "INFO: running local test with " << peers.size() <<
-		" participants" << std::endl;
-	// open pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if (pipe(pipefd[i][j]) < 0)
-				perror("ERROR: dkg-sign (pipe)");
-			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("ERROR: dkg-sign (pipe)");
-		}
 	}
-	
-	// start childs
-	for (size_t i = 0; i < peers.size(); i++)
-		fork_instance(i);
-
-	// sleep for five seconds
-	sleep(5);
-	
-	// wait for childs and close pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		int wstatus = 0;
-		if (opt_verbose)
-			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
-		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("ERROR: dkg-sign (waitpid)");
-		if (!WIFEXITED(wstatus))
-		{
-			std::cerr << "ERROR: protocol instance ";
-			if (WIFSIGNALED(wstatus))
-			{
-				std::cerr << pid[i] << " terminated by signal " <<
-					WTERMSIG(wstatus) << std::endl;
-			}
-			if (WCOREDUMP(wstatus))
-				std::cerr << pid[i] << " dumped core" << std::endl;
-			ret = -1; // fatal error
-		}
-		else if (WIFEXITED(wstatus))
-		{
-			if (opt_verbose)
-			{
-				std::cerr << "INFO: protocol instance " << pid[i] <<
-					" terminated with exit status " << WEXITSTATUS(wstatus) <<
-					std::endl;
-			}
-			if (WEXITSTATUS(wstatus))
-				ret = -2; // error
-		}
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-sign (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
-				(close(broadcast_pipefd[i][j][1]) < 0))
-			{
-				perror("ERROR: dkg-sign (close)");
-			}
-		}
-	}
-	
+	if (should_unlock)
+		unlock_memory();
 	return ret;
 }
 
+
