@@ -73,7 +73,7 @@ std::string						passwords, hostname, port, yfilename;
 
 int 							opt_verbose = 0;
 bool							opt_binary = false, opt_E = false;
-bool							opt_weak = false;
+bool							opt_weak = false, opt_s = false;
 char							*opt_ifilename = NULL;
 char							*opt_ofilename = NULL;
 char							*opt_passwords = NULL;
@@ -615,6 +615,62 @@ bool check_esk
 	return true;
 }
 
+gcry_error_t decrypt_kek
+	(const tmcg_openpgp_octets_t &kek, const tmcg_openpgp_skalgo_t algo,
+	 const tmcg_openpgp_secure_octets_t &key, tmcg_openpgp_secure_octets_t &out)
+{
+	gcry_error_t ret = 0;
+	size_t bs = CallasDonnerhackeFinneyShawThayerRFC4880::
+		AlgorithmIVLength(algo); // get block size of algorithm
+	size_t ks = CallasDonnerhackeFinneyShawThayerRFC4880::
+		AlgorithmKeyLength(algo); // get key size of algorithm
+	if ((bs == 0) || (ks == 0))
+		return gcry_error(GPG_ERR_CIPHER_ALGO); // error: bad algorithm
+	size_t buflen = (kek.size() >= key.size()) ? kek.size() : key.size();
+	unsigned char *buf = (unsigned char*)gcry_malloc_secure(buflen);
+	if (buf == NULL)
+		return gcry_error(GPG_ERR_RESOURCE_LIMIT); // cannot alloc secure memory
+	gcry_cipher_hd_t hd;
+	ret = gcry_cipher_open(&hd, CallasDonnerhackeFinneyShawThayerRFC4880::
+		AlgorithmSymGCRY(algo), GCRY_CIPHER_MODE_CFB, GCRY_CIPHER_SECURE);
+	if (ret)
+	{
+		gcry_free(buf);
+		return ret;
+	}
+	for (size_t i = 0; i < key.size(); i++)
+		buf[i] = key[i];
+	ret = gcry_cipher_setkey(hd, buf, key.size());
+	if (ret)
+	{
+		gcry_free(buf);
+		gcry_cipher_close(hd);
+		return ret;
+	}
+	// set "an IV of all zeros" [RFC 4880]
+	ret = gcry_cipher_setiv(hd, NULL, 0);
+	if (ret)
+	{
+		gcry_free(buf);
+		gcry_cipher_close(hd);
+		return ret;
+	}
+	for (size_t i = 0; i < kek.size(); i++)
+		buf[i] = kek[i];
+	ret = gcry_cipher_decrypt(hd, buf, kek.size(), NULL, 0);
+	if (ret)
+	{
+		gcry_free(buf);
+		gcry_cipher_close(hd);
+		return ret;
+	}
+	for (size_t i = 0; i < kek.size(); i++)
+		out.push_back(buf[i]);
+	gcry_free(buf);
+	gcry_cipher_close(hd);
+	return ret;
+}
+
 bool decompress_libz
 	(const TMCG_OpenPGP_Message* msg, tmcg_openpgp_octets_t &infmsg)
 {
@@ -631,13 +687,13 @@ bool decompress_libz
 	if (zlibVersion()[0] != myZlibVersion[0])
 	{
 		if (opt_verbose > 1)
-			std::cerr << "ERROR: incompatible zlib version" << std::endl;
+			std::cerr << "ERROR: incompatible zlib versions found" << std::endl;
 		return false;
 	}
 	else if (std::strcmp(zlibVersion(), ZLIB_VERSION) != 0)
 	{
 		if (opt_verbose > 1)
-			std::cerr << "WARNING: different zlib version" << std::endl;
+			std::cerr << "WARNING: different zlib versions found" << std::endl;
 	}
 	switch (msg->compalgo)
 	{
@@ -648,7 +704,7 @@ bool decompress_libz
 			rc = inflateInit(&zs);
 			break;
 		default:
-			if (opt_verbose > 1)
+			if (opt_verbose)
 				std::cerr << "ERROR: compression algorithm " <<
 					(int)msg->compalgo << " is not supported" << std::endl;
 			return false;
@@ -680,13 +736,22 @@ bool decompress_libz
 			}
 			zs.avail_in = zlen;
 			zs.next_in = zin;
+			for (size_t i = 0; i < 4; i++)
+			{
+				if ((zs.avail_in < sizeof(zin)) &&
+					(msg->compalgo == TMCG_OPENPGP_COMPALGO_ZIP))
+				{
+					if (opt_verbose > 1)
+					{
+						std::cerr << "INFO: fill with 0xFF at zs.avail_in = " <<
+							zs.avail_in << std::endl;
+					}
+					zin[zs.avail_in] = 0xFF; // dummy byte to fake zlib header
+					zs.avail_in++;
+				}
+			}
 		}
-		if ((zs.avail_in < sizeof(zin)) &&
-			(msg->compalgo == TMCG_OPENPGP_COMPALGO_ZIP))
-		{
-			zin[zs.avail_in] = 0xFF; // dummy character to fake zlib structure
-			zs.avail_in++;
-		}
+
 		memset(zout, 0, sizeof(zout));
 		zs.avail_out = sizeof(zout);
 		zs.next_out = zout;
@@ -780,112 +845,120 @@ bool decompress_libbz
 #endif
 
 bool decrypt_message
-	(const tmcg_openpgp_secure_octets_t seskey, const TMCG_OpenPGP_Keyring *ring,
+	(const tmcg_openpgp_secure_octets_t key, const TMCG_OpenPGP_Keyring *ring,
 	 TMCG_OpenPGP_Message *msg, tmcg_openpgp_octets_t &content)
 {
-	tmcg_openpgp_octets_t decmsg, infmsg;
-	if (!msg->Decrypt(seskey, opt_verbose, decmsg))
+	tmcg_openpgp_octets_t decmsg;
+	if (!opt_s)
 	{
-		std::cerr << "ERROR: message decryption failed" << std::endl;
-		return false;
+		if (!msg->Decrypt(key, opt_verbose, decmsg))
+		{
+			std::cerr << "ERROR: message decryption failed" << std::endl;
+			return false;
+		}
+		if (!CallasDonnerhackeFinneyShawThayerRFC4880::MessageParse(decmsg,
+			opt_verbose, msg))
+		{
+			std::cerr << "ERROR: message parsing failed" << std::endl;
+			return false;
+		}
 	}
-	if (!CallasDonnerhackeFinneyShawThayerRFC4880::MessageParse(decmsg,
-		opt_verbose, msg))
+	// handle compressed message
+	tmcg_openpgp_octets_t infmsg;
+	if ((msg->compressed_data).size() != 0)
 	{
-		std::cerr << "ERROR: message parsing failed" << std::endl;
+		bool decompress_ok = false;
+		switch (msg->compalgo)
+		{
+			case TMCG_OPENPGP_COMPALGO_UNCOMPRESSED:
+				for (size_t i = 0; i < (msg->compressed_data).size(); i++)
+					infmsg.push_back(msg->compressed_data[i]);
+				decompress_ok = true; // no compression
+				break;
+			case TMCG_OPENPGP_COMPALGO_ZIP:
+			case TMCG_OPENPGP_COMPALGO_ZLIB:
+				decompress_ok = decompress_libz(msg, infmsg);
+				break;
+#ifdef LIBBZ
+			case TMCG_OPENPGP_COMPALGO_BZIP2:
+				decompress_ok = decompress_libbz(msg, infmsg);
+				break;
+#endif
+			default:
+				if (opt_verbose > 1)
+				{
+					std::cerr << "WARNING: compression algorithm " <<
+						(int)msg->compalgo << " is not supported" <<
+						std::endl;
+				}
+		}
+		if (!decompress_ok)
+		{
+			std::cerr << "ERROR: decompression failed" << std::endl;
+			return false;
+		}
+		if (!CallasDonnerhackeFinneyShawThayerRFC4880::MessageParse(infmsg,
+			opt_verbose, msg))
+		{
+			std::cerr << "ERROR: message parsing failed" << std::endl;
+			return false;
+		}
+	}
+	// handle decompressed message
+	if ((msg->literal_data).size() == 0)
+	{
+		std::cerr << "ERROR: no literal data in decrypted message" <<
+			std::endl;
 		return false;
 	}
 	else
 	{
-		// handle compressed message
-		if ((msg->compressed_data).size() != 0)
+		// verify included signatures based on keys from keyring
+		if ((opt_k != NULL) && ((msg->signatures).size() > 0))
 		{
-			bool decompress_ok = false;
-			switch (msg->compalgo)
+			bool vf = true, vf_one = false;
+			for (size_t i = 0; i < (msg->signatures).size(); i++)
 			{
-				case TMCG_OPENPGP_COMPALGO_UNCOMPRESSED:
-					for (size_t i = 0; i < (msg->compressed_data).size(); i++)
-						infmsg.push_back(msg->compressed_data[i]);
-					decompress_ok = true; // no compression
-					break;
-				case TMCG_OPENPGP_COMPALGO_ZIP:
-				case TMCG_OPENPGP_COMPALGO_ZLIB:
-					decompress_ok = decompress_libz(msg, infmsg);
-					break;
-#ifdef LIBBZ
-				case TMCG_OPENPGP_COMPALGO_BZIP2:
-					decompress_ok = decompress_libbz(msg, infmsg);
-					break;
-#endif
-				default:
-					if (opt_verbose > 1)
-					{
-						std::cerr << "WARNING: compression algorithm " <<
-							(int)msg->compalgo << " is not supported" <<
-							std::endl;
-					}
-			}
-			if (!decompress_ok)
-			{
-				std::cerr << "ERROR: decompression failed" << std::endl;
-				return false;
-			}
-			if (!CallasDonnerhackeFinneyShawThayerRFC4880::MessageParse(infmsg,
-				opt_verbose, msg))
-			{
-				std::cerr << "ERROR: message parsing failed" << std::endl;
-				return false;
-			}
-		}
-		// handle decompressed message
-		if ((msg->literal_data).size() == 0)
-		{
-			std::cerr << "ERROR: no literal data in decrypted message" <<
-				std::endl;
-			return false;
-		}
-		else
-		{
-			// verify included signatures based on keys from keyring
-			if ((opt_k != NULL) && ((msg->signatures).size() > 0))
-			{
-				bool vf = true;
-				for (size_t i = 0; i < (msg->signatures).size(); i++)
+				const TMCG_OpenPGP_Signature *sig = msg->signatures[i];
+				std::string ak;
+				if (get_key_by_signature(ring, sig, opt_verbose, ak))
 				{
-					const TMCG_OpenPGP_Signature *sig = msg->signatures[i];
-					std::string ak;
-					if (get_key_by_signature(ring, sig, opt_verbose, ak))
+					if (!verify_signature(msg->literal_data, ak, sig, ring,
+						opt_verbose, opt_weak))
 					{
-						if (!verify_signature(msg->literal_data, ak, sig,
-							ring, opt_verbose, opt_weak))
-						{
-							vf = false;
-							std::cerr << "WARNING: verification of" <<
+						vf = false;
+						std::cerr << "WARNING: verification of included" <<
 							" signature #" << i << " failed" << std::endl;
-						}	
 					}
 					else
-					{
-						std::cerr << "WARNING: cannot verify" <<
-							" signature #" << i << " due to missing" <<
-							" public key" << std::endl;
-					}
+						vf_one = true;
 				}
-				if (!vf)
+				else
 				{
-					std::cerr << "ERROR: verification of included" <<
-						" signature(s) failed" << std::endl;
-					return false;
+					std::cerr << "WARNING: cannot verify included signature" <<
+						" #" << i << " due to missing public key" << std::endl;
 				}
 			}
-			// copy the content of literal data packet
-			content.insert(content.end(), (msg->literal_data).begin(),
-				(msg->literal_data).end());
+			if (!vf)
+			{
+				std::cerr << "ERROR: verification of included signature(s)" <<
+					" failed" << std::endl;
+				return false;
+			}
+			if (opt_s && !vf_one)
+			{
+				std::cerr << "ERROR: none of the included signature(s)" <<
+					" is valid" << std::endl;
+				return false;
+			}
 		}
-		if (msg->filename == "_CONSOLE")
-			std::cerr << "INFO: sender requested \"for-your-eyes-only\"" <<
-				std::endl;
+		// copy the content of literal data packet
+		content.insert(content.end(), (msg->literal_data).begin(),
+			(msg->literal_data).end());
 	}
+	if (msg->filename == "_CONSOLE")
+		std::cerr << "INFO: sender requested \"for-your-eyes-only\"" <<
+			std::endl;
 	return true;
 }
 
@@ -925,8 +998,11 @@ void run_instance
 	bool parse_ok;
 	if (opt_k)
 	{
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
-			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+			PublicKeyringParse(armored_pubring, opt_verbose_ring, ring);
 		if (!parse_ok)
 		{
 			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
@@ -1122,7 +1198,7 @@ void run_instance
 		delete prv;
 		exit(-1);
 	}
-	if (msg->encrypted_message.size() == 0)
+	if (!opt_s && (msg->encrypted_message.size() == 0))
 	{
 		std::cerr << "ERROR: no encrypted data found" << std::endl;
 		delete msg;
@@ -1181,7 +1257,7 @@ void run_instance
 			}
 		}
 	}
-	if ((esks.size() == 0) && (msg->SKESKs.size() == 0))
+	if ((esks.size() == 0) && (msg->SKESKs.size() == 0) && !opt_s)
 	{
 		std::cerr << "ERROR: no admissible encrypted session key found" <<
 			std::endl;
@@ -1622,25 +1698,32 @@ void run_instance
 			for (size_t i = 0; i < msg->SKESKs.size(); i++)
 			{
 				const TMCG_OpenPGP_SKESK *esk = msg->SKESKs[i];
-				tmcg_openpgp_secure_octets_t esk_seskey;
+				tmcg_openpgp_secure_octets_t kek;
+				size_t klen = CallasDonnerhackeFinneyShawThayerRFC4880::
+					AlgorithmKeyLength(esk->skalgo);
+				if (opt_verbose > 2)
+				{
+					std::cerr << "INFO: s2k_salt = " << std::hex;
+					for (size_t j = 0; j < (esk->s2k_salt).size(); j++)
+						std::cerr << (int)esk->s2k_salt[j] << " ";
+					std::cerr << std::dec << std::endl;
+				}
 				switch (esk->s2k_type)
 				{
 					case TMCG_OPENPGP_STRINGTOKEY_SIMPLE:
+						if (opt_verbose)
+							std::cerr << "WARNING: S2K specifier not" <<
+								" supported; skip SKESK" << std::endl;
+						break;
 					case TMCG_OPENPGP_STRINGTOKEY_SALTED:
-						CallasDonnerhackeFinneyShawThayerRFC4880::S2KCompute(
-								esk->s2k_hashalgo,
-								CallasDonnerhackeFinneyShawThayerRFC4880::
-									AlgorithmKeyLength(esk->skalgo),
-								esk_passphrase, esk->s2k_salt, false,
-								esk->s2k_count, esk_seskey);
+						CallasDonnerhackeFinneyShawThayerRFC4880::
+							S2KCompute(esk->s2k_hashalgo, klen, esk_passphrase,
+								esk->s2k_salt, false, esk->s2k_count, kek);
 						break;
 					case TMCG_OPENPGP_STRINGTOKEY_ITERATED:
-						CallasDonnerhackeFinneyShawThayerRFC4880::S2KCompute(
-								esk->s2k_hashalgo,
-								CallasDonnerhackeFinneyShawThayerRFC4880::
-									AlgorithmKeyLength(esk->skalgo),
-								esk_passphrase, esk->s2k_salt, true,
-								esk->s2k_count, esk_seskey);
+						CallasDonnerhackeFinneyShawThayerRFC4880::
+							S2KCompute(esk->s2k_hashalgo, klen, esk_passphrase,
+								esk->s2k_salt, true, esk->s2k_count, kek);
 						break;
 					default:
 						if (opt_verbose)
@@ -1648,16 +1731,24 @@ void run_instance
 								" supported; skip SKESK" << std::endl;
 						break;
 				}
-				seskey.clear();
-				if (esk->encrypted_key.size() != 0)
+				if (opt_verbose > 2)
 				{
-					tmcg_openpgp_octets_t decrypted_key, prefix;
+					std::cerr << "INFO: kek.size() = " << kek.size() <<
+						std::endl;
+					std::cerr << "INFO: kek = " << std::hex;
+					for (size_t j = 0; j < kek.size(); j++)
+						std::cerr << (int)kek[j] << " ";
+					std::cerr << std::dec << std::endl;
+				}
+				seskey.clear();
+				if (esk->encrypted_key.size() > 0)
+				{
+					tmcg_openpgp_octets_t decrypted_key;
 					gcry_error_t ret = 0;
 					if (esk->aeadalgo == 0)
 					{
-						ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-							SymmetricDecrypt(esk->encrypted_key, esk_seskey,
-								prefix, false, esk->skalgo, decrypted_key);
+						ret = decrypt_kek(esk->encrypted_key, esk->skalgo, kek,
+							seskey);
 					}
 					else
 					{
@@ -1667,7 +1758,7 @@ void run_instance
 						ad.push_back(esk->skalgo); // cipher algorithm octet
 						ad.push_back(esk->aeadalgo); // AEAD algorithm octet
 						ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-							SymmetricDecryptAEAD(esk->encrypted_key, esk_seskey,
+							SymmetricDecryptAEAD(esk->encrypted_key, kek,
 								esk->skalgo, esk->aeadalgo, 0, esk->iv, ad,
 								opt_verbose, decrypted_key);
 					}
@@ -1688,8 +1779,8 @@ void run_instance
 				else
 				{
 					seskey.push_back(esk->skalgo);
-					for (size_t j = 0; j < esk_seskey.size(); j++)
-						seskey.push_back(esk_seskey[j]);
+					for (size_t j = 0; j < kek.size(); j++)
+						seskey.push_back(kek[j]);
 				}
 				// quick check, whether decryption of session key was successful
 				tmcg_openpgp_octets_t tmpmsg;				
@@ -1710,7 +1801,7 @@ void run_instance
 				exit(-1);
 			}
 		}
-		else
+		else if (!opt_s)
 		{
 			std::cerr << "ERROR: every PKESK decryption failed" << std::endl;
 			delete msg;
@@ -1780,6 +1871,7 @@ int gnunet_opt_verbose = 0;
 int gnunet_opt_weak = 0;
 int gnunet_opt_binary = 0;
 int gnunet_opt_E = 0;
+int gnunet_opt_s = 0;
 #endif
 
 void fork_instance
@@ -1880,6 +1972,11 @@ int main
 			"exchanged passwords to protect private and broadcast channels",
 			&gnunet_opt_passwords
 		),
+		GNUNET_GETOPT_option_flag('s',
+			"signed-only",
+			"allow signed-only message",
+			&gnunet_opt_s
+		),
 		GNUNET_GETOPT_option_version(version),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
@@ -2009,7 +2106,8 @@ int main
 		else if ((arg.find("--") == 0) || (arg.find("-b") == 0) ||
 			(arg.find("-v") == 0) || (arg.find("-h") == 0) ||
 			(arg.find("-n") == 0) || (arg.find("-V") == 0) ||
-			(arg.find("-E") == 0) || (arg.find("-K") == 0))
+			(arg.find("-E") == 0) || (arg.find("-K") == 0) ||
+			(arg.find("-s") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -2040,6 +2138,8 @@ int main
 					" built-in TCP/IP message exchange service" << std::endl;
 				std::cout << "  -P STRING              exchanged passwords" <<
 					" to protect private and broadcast channels" << std::endl;
+				std::cout << "  -s, --signed-only      allow signed-only" <<
+					" message" << std::endl;
 				std::cout << "  -v, --version          print the version" <<
 					" number" << std::endl;
 				std::cout << "  -V, --verbose          turn on verbose" <<
@@ -2057,6 +2157,8 @@ int main
 				opt_E = true;
 			if ((arg.find("-K") == 0) || (arg.find("--weak") == 0))
 				opt_weak = true;
+			if ((arg.find("-s") == 0) || (arg.find("--signed-only") == 0))
+				opt_s = true;
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
 			{
 #ifndef GNUNET
@@ -2237,8 +2339,11 @@ int main
 		bool parse_ok;
 		if (opt_k)
 		{
+			int opt_verbose_ring = opt_verbose;
+			if (opt_verbose_ring > 0)
+				opt_verbose_ring--;
 			parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
-				PublicKeyringParse(armored_pubring, opt_verbose, ring);
+				PublicKeyringParse(armored_pubring, opt_verbose_ring, ring);
 			if (!parse_ok)
 			{
 				std::cerr << "WARNING: cannot use the given keyring" <<
@@ -2590,6 +2695,11 @@ int main
 				"exchanged passwords to protect private and broadcast channels",
 				&gnunet_opt_passwords
 			),
+			GNUNET_GETOPT_option_flag('s',
+				"signed-only",
+				"allow signed-only message",
+				&gnunet_opt_s
+			),
 			GNUNET_GETOPT_option_flag('V',
 				"verbose",
 				"turn on verbose output",
