@@ -64,6 +64,7 @@ bool							instance_forked = false;
 
 tmcg_openpgp_secure_string_t	passphrase;
 std::string						ifilename, ofilename, kfilename;
+std::string						Kfilename, fingerprint;
 std::string						passwords, hostname, port, URI, u, yfilename;
 
 int 							opt_verbose = 0;
@@ -74,10 +75,12 @@ char							*opt_hostname = NULL;
 char							*opt_URI = NULL;
 char							*opt_u = NULL;
 char							*opt_k = NULL;
+char							*opt_K = NULL;
+char							*opt_fingerprint = NULL;
 char							*opt_y = NULL;
 unsigned long int				opt_e = 0, opt_p = 55000, opt_W = 5;
-bool							opt_r = false;
 bool							opt_1 = false, opt_2 = false, opt_3 = false;
+bool							opt_a = false, opt_r = false;
 
 void run_instance
 	(size_t whoami, const time_t sigtime, const time_t sigexptime,
@@ -103,9 +106,43 @@ void run_instance
 		exit(-1);
 
 	// read the public key
+	bool parse_ok;
 	std::string armored_pubkey;
-	if (!read_key_file(opt_ifilename, armored_pubkey))
-		exit(-1);
+	if (opt_ifilename != NULL)
+	{
+		if (!read_key_file(opt_ifilename, armored_pubkey))
+			exit(-1);
+	}
+	else if (opt_K != NULL)
+	{
+		std::string armored_certring;
+		if (!read_key_file(Kfilename, armored_certring))
+			exit(-1);
+		TMCG_OpenPGP_Keyring *certring = NULL;
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PublicKeyringParse(armored_certring, opt_verbose_ring, certring);
+		if (!parse_ok)
+		{
+			std::cerr << "ERROR: cannot use the given keyring" << std::endl;
+			exit(-1);
+		}
+		if (!get_key_by_fingerprint(certring, fingerprint, opt_verbose,
+			armored_pubkey))
+		{
+			if (!get_key_by_keyid(certring, fingerprint, opt_verbose,
+				armored_pubkey))
+			{
+				std::cerr << "ERROR: public key not found in keyring" <<
+					std::endl;
+				delete certring;
+				exit(-1);
+			}
+		}
+		delete certring;
+	}
 
 	// read the keyring
 	std::string armored_pubring;
@@ -118,11 +155,13 @@ void run_instance
 	// parse the keyring, the private key and corresponding signatures
 	TMCG_OpenPGP_Prvkey *prv = NULL;
 	TMCG_OpenPGP_Keyring *ring = NULL;
-	bool parse_ok;
 	if (opt_k)
 	{
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
-			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+			PublicKeyringParse(armored_pubring, opt_verbose_ring, ring);
 		if (!parse_ok)
 		{
 			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
@@ -330,38 +369,41 @@ void run_instance
 	// prepare the fingerprint, the trailer, and the accumulator of the
 	// certification (revocation) signatures
 	std::string fpr;
-	tmcg_openpgp_octets_t trailer, acc, issuer;
+	tmcg_openpgp_octets_t trailer, acc;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintComputePretty(primary->pub_hashing, fpr);
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(prv->pub->pub_hashing, issuer);
+		FingerprintConvertPretty(primary->fingerprint, fpr);
 	if (opt_y == NULL)
 	{
 		if (opt_r)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CERTIFICATION_REVOCATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_1)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_PERSONA_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_2)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CASUAL_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_3)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_GENERIC_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 	}
 	else
 	{
@@ -369,31 +411,37 @@ void run_instance
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CERTIFICATION_REVOCATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_1)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_PERSONA_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_2)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CASUAL_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_3)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_GENERIC_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 	}
 	acc.insert(acc.end(), primary->packet.begin(), primary->packet.end());
 
-	// loop through all or selected valid user IDs	
+	// loop through all or selected valid user IDs
+	bool anything_signed = false;
 	for (size_t j = 0; j < primary->userids.size(); j++)
 	{
 		// user ID not selected?
@@ -413,6 +461,17 @@ void run_instance
 			std::cerr << "INFO: going to sign user ID \"" <<
 				primary->userids[j]->userid_sanitized << "\" of key with" <<
 				" fingerprint [ " << fpr << "]" << std::endl;
+		if (opt_a)
+		{
+			std::string p = "Please confirm operation by entering the string";
+			if (!check_confirmation(p))
+			{
+				std::cerr << "WARNING: operation has been cancelled" <<
+					std::endl;
+				continue;
+			}
+		}
+		anything_signed = true;
 
 		// sign the hash value
 		tmcg_openpgp_octets_t sig;
@@ -480,6 +539,13 @@ void run_instance
 	delete primary;
 	delete prv;
 
+	// check the result
+	if (!anything_signed)
+	{
+		std::cerr << "ERROR: no user ID selected and signed" << std::endl;
+		exit(-2);
+	}
+
 	// output the result
 	std::string signedkey;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
@@ -501,6 +567,8 @@ char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 char *gnunet_opt_u = NULL;
 char *gnunet_opt_URI = NULL;
+char *gnunet_opt_K = NULL;
+char *gnunet_opt_fingerprint = NULL;
 char *gnunet_opt_k = NULL;
 char *gnunet_opt_y = NULL;
 unsigned int gnunet_opt_sigexptime = 0;
@@ -508,8 +576,8 @@ unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
 unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_verbose = 0;
-int gnunet_opt_r = 0;
 int gnunet_opt_1 = 0, gnunet_opt_2 = 0, gnunet_opt_3 = 0;
+int gnunet_opt_a = 0, gnunet_opt_r = 0;
 #endif
 
 void fork_instance
@@ -569,12 +637,23 @@ int main
 			"issuer has done substantial verification of the claim of identity",
 			&gnunet_opt_3
 		),
+		GNUNET_GETOPT_option_flag('a',
+			"ask",
+			"require confirmation from STDIN for each signature",
+			&gnunet_opt_a
+		),
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
 			"INTEGER",
 			"expiration time of generated signature in seconds",
 			&gnunet_opt_sigexptime
 		),
+		GNUNET_GETOPT_option_string('f',
+			"fingerprint",
+			"STRING",
+			"fingerprint of the public key for certification",
+			&gnunet_opt_fingerprint
+		),
 		GNUNET_GETOPT_option_string('H',
 			"hostname",
 			"STRING",
@@ -587,6 +666,12 @@ int main
 			"create certification signature on key resp. user ID from FILENAME",
 			&gnunet_opt_ifilename
 		),
+		GNUNET_GETOPT_option_string('K',
+			"keys",
+			"FILENAME",
+			"select public key for certification from keyring FILENAME",
+			&gnunet_opt_K
+		),
 		GNUNET_GETOPT_option_string('k',
 			"keyring",
 			"FILENAME",
@@ -685,6 +770,10 @@ int main
 		opt_URI = gnunet_opt_URI;
 	if (gnunet_opt_u != NULL)
 		opt_u = gnunet_opt_u;
+	if (gnunet_opt_K != NULL)
+		opt_K = gnunet_opt_K;
+	if (gnunet_opt_fingerprint != NULL)
+		opt_fingerprint = gnunet_opt_fingerprint;
 	if (gnunet_opt_k != NULL)
 		opt_k = gnunet_opt_k;
 	if (gnunet_opt_passwords != NULL)
@@ -697,6 +786,10 @@ int main
 		URI = gnunet_opt_URI; // get policy URI from GNUnet options
 	if (gnunet_opt_u != NULL)
 		u = gnunet_opt_u; // get policy URI from GNUnet options
+	if (gnunet_opt_K != NULL)
+		Kfilename = gnunet_opt_K; // get keyring from GNUnet options
+	if (gnunet_opt_fingerprint != NULL)
+		fingerprint = gnunet_opt_fingerprint; // get fingerprint from GNUnet options
 	if (gnunet_opt_k != NULL)
 		kfilename = gnunet_opt_k; // get keyring from GNUnet options
 	if (gnunet_opt_y != NULL)
@@ -715,9 +808,16 @@ int main
 		    (arg.find("-e") == 0) || (arg.find("-x") == 0) ||
 			(arg.find("-P") == 0) || (arg.find("-H") == 0) ||
 		    (arg.find("-u") == 0) || (arg.find("-U") == 0) ||
-			(arg.find("-k") == 0) || (arg.find("-y") == 0))
+			(arg.find("-K") == 0) || (arg.find("-k") == 0) ||
+			(arg.find("-f") == 0) || (arg.find("-y") == 0))
 		{
 			size_t idx = ++i;
+			if ((arg.find("-f") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_fingerprint == NULL))
+			{
+				fingerprint = argv[i+1];
+				opt_fingerprint = (char*)fingerprint.c_str();
+			}
 			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) &&
 				(opt_ifilename == NULL))
 			{
@@ -754,6 +854,12 @@ int main
 				URI = argv[i+1];
 				opt_URI = (char*)URI.c_str();
 			}
+			if ((arg.find("-K") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_K == NULL))
+			{
+				Kfilename = argv[i+1];
+				opt_K = (char*)Kfilename.c_str();
+			}
 			if ((arg.find("-k") == 0) && (idx < (size_t)(argc - 1)) &&
 				(opt_k == NULL))
 			{
@@ -783,10 +889,11 @@ int main
 			}
 			continue;
 		}
-		else if ((arg.find("--") == 0) || (arg.find("-r") == 0) ||
-			(arg.find("-v") == 0) || (arg.find("-h") == 0) ||
-			(arg.find("-V") == 0) || (arg.find("-1") == 0) ||
-			(arg.find("-2") == 0) || (arg.find("-3") == 0))
+		else if ((arg.find("--") == 0) || (arg.find("-a") == 0) ||
+			(arg.find("-r") == 0) || (arg.find("-v") == 0) ||
+			(arg.find("-h") == 0) || (arg.find("-V") == 0) ||
+			(arg.find("-1") == 0) || (arg.find("-2") == 0) ||
+			(arg.find("-3") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -795,19 +902,25 @@ int main
 				std::cout << about << std::endl;
 				std::cout << "Arguments mandatory for long options are also" <<
 					" mandatory for short options." << std::endl;
-				std::cout << "  -h, --help       print this help" << std::endl;
 				std::cout << "  -1, --one        issuer has not done any" <<
 					" verification of claim of identity" << std::endl;
 				std::cout << "  -2, --two        issuer has done some casual" <<
 					" verification of claim of identity" << std::endl;
 				std::cout << "  -3, --three      issuer has done substantial" <<
 					" verification of claim of identity" << std::endl;
+				std::cout << "  -a, --ask        require confirmation from" <<
+					" STDIN for each signature" << std::endl;
 				std::cout << "  -e INTEGER       expiration time of" <<
 					" generated signatures in seconds" << std::endl;
+				std::cout << "  -f STRING        fingerprint of the public" <<
+					" key for certification" << std::endl;
 				std::cout << "  -H STRING        hostname (e.g. onion" <<
 					" address) of this peer within PEERS" << std::endl;
+				std::cout << "  -h, --help       print this help" << std::endl;
 				std::cout << "  -i FILENAME      create certification" <<
 					" signatures on key from FILENAME" << std::endl;
+				std::cout << "  -K FILENAME      select public key for" <<
+					" certification from keyring FILENAME" << std::endl;
 				std::cout << "  -k FILENAME      use keyring FILENAME" <<
 					" containing external revocation keys" << std::endl;
 				std::cout << "  -o FILENAME      write key with" <<
@@ -834,6 +947,8 @@ int main
 #endif
 				return 0; // not continue
 			}
+			if ((arg.find("-a") == 0) || (arg.find("--ask") == 0))
+				opt_a = true; // require confirmation
 			if ((arg.find("-r") == 0) || (arg.find("--revocation") == 0))
 				opt_r = true; // create revocation signature
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
@@ -897,10 +1012,16 @@ int main
 #endif
 
 	// check command line arguments
-	if (opt_ifilename == NULL)
+	if ((opt_ifilename == NULL) && (opt_K == NULL))
 	{
-		std::cerr << "ERROR: option -i required to specify an input file" <<
-			std::endl;
+		std::cerr << "ERROR: option \"-i\" or \"-K\" is required to specify" <<
+			" an input file" << std::endl;
+		return -1;
+	}
+	if ((opt_K != NULL) && (opt_fingerprint == NULL))
+	{
+		std::cerr << "ERROR: option \"-f\" is required to select public key" <<
+			" for certification" << std::endl;
 		return -1;
 	}
 	if ((opt_hostname != NULL) && (opt_passwords == NULL) && (opt_y == NULL))
@@ -985,12 +1106,23 @@ int main
 				"issuer has done substantial verification of the claim of identity",
 				&gnunet_opt_3
 			),
+			GNUNET_GETOPT_option_flag('a',
+				"ask",
+				"require confirmation from STDIN for each signature",
+				&gnunet_opt_a
+			),
 			GNUNET_GETOPT_option_uint('e',
 				"expiration",
 				"INTEGER",
 				"expiration time of generated signature in seconds",
 				&gnunet_opt_sigexptime
 			),
+			GNUNET_GETOPT_option_string('f',
+				"fingerprint",
+				"STRING",
+				"fingerprint of the public key for certification",
+				&gnunet_opt_fingerprint
+			),
 			GNUNET_GETOPT_option_string('H',
 				"hostname",
 				"STRING",
@@ -1003,6 +1135,12 @@ int main
 				"create certification signature on key resp. user ID from FILENAME",
 				&gnunet_opt_ifilename
 			),
+			GNUNET_GETOPT_option_string('K',
+				"keys",
+				"FILENAME",
+				"select public key for certification from keyring FILENAME",
+				&gnunet_opt_K
+			),
 			GNUNET_GETOPT_option_string('k',
 				"keyring",
 				"FILENAME",
