@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -131,11 +131,13 @@ void tcpip_init
 }
 
 void tcpip_bindports
-	(const uint16_t start, const bool broadcast)
+	(const uint16_t stpo, const bool broadcast)
 {
 	if (opt_verbose > 2)
-		std::cerr << "INFO: tcpip_bindports(" << start << ", " <<
+		std::cerr << "INFO: tcpip_bindports(" << stpo << ", " <<
 			(broadcast ? "true" : "false") << ") called" << std::endl;
+	if (opt_verbose > 2)
+		std::cerr << "INFO: FD_SETSIZE = " << FD_SETSIZE << std::endl;
 	uint16_t peers_size = 0;
 	if (peers.size() <= DKGPG_MAX_N)
 	{
@@ -149,7 +151,9 @@ void tcpip_bindports
 		exit(-1);
 	}
 	uint16_t peer_offset = tcpip_peer2pipe[tcpip_thispeer] * peers_size;
-	uint16_t local_start = start + peer_offset;
+	if (opt_verbose > 2)
+		std::cerr << "INFO: peer_offset = " << peer_offset << std::endl;
+	uint16_t local_start = stpo + peer_offset;
 	uint16_t local_end = local_start + peers_size;
 	size_t i = 0;
 	if (broadcast)
@@ -159,8 +163,8 @@ void tcpip_bindports
 	}
 	for (uint16_t port = local_start; port < local_end; port++, i++)
 	{
-		struct addrinfo hints = { 0 }, *res, *rp;
-		hints.ai_family = AF_INET;
+		struct addrinfo hints = { 0, 0, 0, 0, 0, 0, 0, 0 }, *res, *rp;
+		hints.ai_family = AF_UNSPEC; // AF_INET; FIXME: resolving IPv4-only does not work
 		hints.ai_socktype = SOCK_STREAM;
 		hints.ai_flags = AI_PASSIVE | AI_NUMERICSERV | AI_ADDRCONFIG;
 		std::stringstream ports;
@@ -246,10 +250,10 @@ void tcpip_bindports
 }
 
 size_t tcpip_connect
-	(const uint16_t start, const bool broadcast)
+	(const uint16_t stpo, const bool broadcast)
 {
 	if (opt_verbose > 2)
-		std::cerr << "INFO: tcpip_connect(" << start << ", " <<
+		std::cerr << "INFO: tcpip_connect(" << stpo << ", " <<
 			(broadcast ? "true" : "false") << ") called" << std::endl;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -269,10 +273,10 @@ size_t tcpip_connect
 				exit(-1);
 			}
 			uint16_t peer_offset = (uint16_t)tcpip_peer2pipe[tcpip_thispeer];
-			uint16_t port = start + (i * peers_size) + peer_offset;
+			uint16_t port = stpo + (i * peers_size) + peer_offset;
 			int ret;
-			struct addrinfo hints = { 0 }, *res, *rp;
-			hints.ai_family = AF_INET;
+			struct addrinfo hints = { 0, 0, 0, 0, 0, 0, 0, 0 }, *res, *rp;
+			hints.ai_family = AF_UNSPEC; // AF_INET; FIXME: resolving IPv4-only does not work
 			hints.ai_socktype = SOCK_STREAM;
 			hints.ai_flags = AI_NUMERICSERV | AI_ADDRCONFIG;
 			if (broadcast)
@@ -360,6 +364,8 @@ void tcpip_accept
 {
 	if (opt_verbose > 2)
 		std::cerr << "INFO: tcpip_accept(...) called" << std::endl;
+	if (opt_verbose > 2)
+		std::cerr << "INFO: FD_SETSIZE = " << FD_SETSIZE << std::endl;
 	while ((tcpip_pipe2socket_in.size() < peers.size()) ||
 		(tcpip_broadcast_pipe2socket_in.size() < peers.size()))
 	{
@@ -370,16 +376,38 @@ void tcpip_accept
 		for (tcpip_mci_t pi = tcpip_pipe2socket.begin();
 			pi != tcpip_pipe2socket.end(); ++pi)
 		{
-			FD_SET(pi->second, &rfds);
-			if (pi->second > maxfd)
-				maxfd = pi->second;
+			if (pi->second < FD_SETSIZE)
+			{
+				FD_SET(pi->second, &rfds);
+				if (pi->second > maxfd)
+					maxfd = pi->second;
+			}
+			else
+			{
+				std::cerr << "ERROR: file descriptor value of internal pipe" <<
+					" exceeds FD_SETSIZE" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
+			}
 		}
 		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket.begin();
 			pi != tcpip_broadcast_pipe2socket.end(); ++pi)
 		{
-			FD_SET(pi->second, &rfds);
-			if (pi->second > maxfd)
-				maxfd = pi->second;
+			if (pi->second < FD_SETSIZE)
+			{
+				FD_SET(pi->second, &rfds);
+				if (pi->second > maxfd)
+					maxfd = pi->second;
+			}
+			else
+			{
+				std::cerr << "ERROR: file descriptor value of internal pipe" <<
+					" exceeds FD_SETSIZE" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
+			}
 		}
 		tv.tv_sec = 1;
 		tv.tv_usec = 0;
@@ -407,8 +435,9 @@ void tcpip_accept
 		{
 			if (FD_ISSET(pi->second, &rfds))
 			{
-				struct sockaddr_storage sin = { 0 };
+				struct sockaddr_storage sin;
 				socklen_t slen = (socklen_t)sizeof(sin);
+				memset(&sin, 0, sizeof(sin));
 				int connfd = accept(pi->second, (struct sockaddr*)&sin, &slen);
 				if (connfd < 0)
 				{
@@ -434,7 +463,7 @@ void tcpip_accept
 					exit(-1);
 				}
 				if (opt_verbose)
-					std::cerr << "INFO: accept connection for P/D/R/S_" <<
+					std::cerr << "INFO: accept connection for p_" <<
 						pi->first << " from address " << ipaddr << std::endl;
 			}
 		}
@@ -443,8 +472,9 @@ void tcpip_accept
 		{
 			if (FD_ISSET(pi->second, &rfds))
 			{
-				struct sockaddr_storage sin = { 0 };
+				struct sockaddr_storage sin;
 				socklen_t slen = (socklen_t)sizeof(sin);
+				memset(&sin, 0, sizeof(sin));
 				int connfd = accept(pi->second, (struct sockaddr*)&sin, &slen);
 				if (connfd < 0)
 				{
@@ -469,14 +499,13 @@ void tcpip_accept
 				}
 				if (opt_verbose)
 					std::cerr << "INFO: accept broadcast connection for" <<
-						" P/D/R/S_" << pi->first << " from address " << 
+						" p_" << pi->first << " from address " << 
 						ipaddr << std::endl;
 			}
 		}
 	}
 }
 
-
 void tcpip_fork
 	()
 {
@@ -544,9 +573,20 @@ int tcpip_io
 		{
 			if (pi->first != thisidx)
 			{
-				FD_SET(pi->second, &rfds);
-				if (pi->second > maxfd)
-					maxfd = pi->second;
+				if (pi->second < FD_SETSIZE)
+				{
+					FD_SET(pi->second, &rfds);
+					if (pi->second > maxfd)
+						maxfd = pi->second;
+				}
+				else
+				{
+					std::cerr << "ERROR: file descriptor value of internal" <<
+						" pipe exceeds FD_SETSIZE" << std::endl;
+					tcpip_close();
+					tcpip_done();
+					exit(-1);
+				}
 			}
 		}
 		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
@@ -554,21 +594,54 @@ int tcpip_io
 		{
 			if (pi->first != thisidx)
 			{
-				FD_SET(pi->second, &rfds);
-				if (pi->second > maxfd)
-					maxfd = pi->second;
+				if (pi->second < FD_SETSIZE)
+				{
+					FD_SET(pi->second, &rfds);
+					if (pi->second > maxfd)
+						maxfd = pi->second;
+				}
+				else
+				{
+					std::cerr << "ERROR: file descriptor value of internal" <<
+						" pipe exceeds FD_SETSIZE" << std::endl;
+					tcpip_close();
+					tcpip_done();
+					exit(-1);
+				}
 			}
 		}
 		for (size_t i = 0; i < peers.size(); i++)
 		{
 			if (i != thisidx)
 			{
-				FD_SET(pipefd[thisidx][i][0], &rfds);
-				if (pipefd[thisidx][i][0] > maxfd)
-					maxfd = pipefd[thisidx][i][0];
-				FD_SET(broadcast_pipefd[thisidx][i][0], &rfds);
-				if (broadcast_pipefd[thisidx][i][0] > maxfd)
-					maxfd = broadcast_pipefd[thisidx][i][0];
+				if (pipefd[thisidx][i][0] < FD_SETSIZE)
+				{
+					FD_SET(pipefd[thisidx][i][0], &rfds);
+					if (pipefd[thisidx][i][0] > maxfd)
+						maxfd = pipefd[thisidx][i][0];
+				}
+				else
+				{
+					std::cerr << "ERROR: file descriptor value of internal" <<
+						" pipe exceeds FD_SETSIZE" << std::endl;
+					tcpip_close();
+					tcpip_done();
+					exit(-1);
+				}
+				if (broadcast_pipefd[thisidx][i][0] < FD_SETSIZE)
+				{
+					FD_SET(broadcast_pipefd[thisidx][i][0], &rfds);
+					if (broadcast_pipefd[thisidx][i][0] > maxfd)
+						maxfd = broadcast_pipefd[thisidx][i][0];
+				}
+				else
+				{
+					std::cerr << "ERROR: file descriptor value of internal" <<
+						" pipe exceeds FD_SETSIZE" << std::endl;
+					tcpip_close();
+					tcpip_done();
+					exit(-1);
+				}
 			}
 		}
 		tv.tv_sec = 1;
@@ -621,7 +694,7 @@ int tcpip_io
 				else if (len == 0)
 				{
 					std::cerr << "WARNING: connection collapsed for" <<
-						" P/D/R/S_" << pi->first << std::endl;
+						" p_" << pi->first << std::endl;
 					tcpip_pipe2socket_out.erase(pi->first);
 					tcpip_pipe2socket_in.erase(pi->first);
 					break;
@@ -630,7 +703,7 @@ int tcpip_io
 				{
 					if (opt_verbose > 1)
 						std::cerr << "INFO: received " << len << " bytes on " <<
-							"connection for P/D/R/S_" << pi->first << std::endl;
+							"connection for p_" << pi->first << std::endl;
 					ssize_t wnum = 0;
 					do
 					{
@@ -699,7 +772,7 @@ int tcpip_io
 				else if (len == 0)
 				{
 					std::cerr << "WARNING: broadcast connection collapsed" <<
-						" for P/D/R/S_" << pi->first << std::endl;
+						" for p_" << pi->first << std::endl;
 					tcpip_broadcast_pipe2socket_out.erase(pi->first);
 					tcpip_broadcast_pipe2socket_in.erase(pi->first);
 					break;
@@ -708,7 +781,7 @@ int tcpip_io
 				{
 					if (opt_verbose > 1)
 						std::cerr << "INFO: received " << len << " bytes on" <<
-							" broadcast connection for P/D/R/S_" <<
+							" broadcast connection for p_" <<
 							pi->first << std::endl;
 					ssize_t wnum = 0;
 					do
@@ -784,7 +857,7 @@ int tcpip_io
 				{
 					if (opt_verbose > 1)
 						std::cerr << "INFO: sending " << len << " bytes on" <<
-							" connection to P/D/R/S_" << i << std::endl;
+							" connection to p_" << i << std::endl;
 					ssize_t wnum = 0;
 					do
 					{
@@ -812,7 +885,7 @@ int tcpip_io
 							else if (errno == ECONNRESET)
 							{
 								std::cerr << "WARNING: connection collapsed" <<
-									" for P/D/R/S_" << i << std::endl;
+									" for p_" << i << std::endl;
 								tcpip_broadcast_pipe2socket_out.erase(i);
 								tcpip_broadcast_pipe2socket_in.erase(i);
 								break;
@@ -834,7 +907,7 @@ int tcpip_io
 				{
 					if (opt_verbose > 1)
 						std::cerr << "INFO: discarding " << len << " bytes" <<
-							" for P/D/R/S_" << i << std::endl;
+							" for p_" << i << std::endl;
 				}
 			}
 			if ((i != thisidx) &&
@@ -870,7 +943,7 @@ int tcpip_io
 				{
 					if (opt_verbose > 1)
 						std::cerr << "INFO: sending " << len << " bytes on" <<
-							" broadcast connection to P/D/R/S_" << i <<
+							" broadcast connection to p_" << i <<
 							std::endl;
 					ssize_t wnum = 0;
 					do
@@ -899,7 +972,7 @@ int tcpip_io
 							else if (errno == ECONNRESET)
 							{
 								std::cerr << "WARNING: broadcast connection" <<
-									" collapsed for P/D/R/S_" << i << std::endl;
+									" collapsed for p_" << i << std::endl;
 								tcpip_broadcast_pipe2socket_out.erase(i);
 								tcpip_broadcast_pipe2socket_in.erase(i);
 								break;
@@ -921,7 +994,7 @@ int tcpip_io
 				{
 					if (opt_verbose > 1)
 						std::cerr << "INFO: discarding " << len << " bytes" <<
-							" for P/D/R/S_" << i << std::endl;
+							" for p_" << i << std::endl;
 				}
 			}
 		}
@@ -977,7 +1050,7 @@ void tcpip_done
 		int thispid = pid[tcpip_peer2pipe[tcpip_thispeer]];
 		if (opt_verbose)
 			std::cerr << "INFO: kill(" << thispid << ", SIGTERM)" << std::endl;
-		if(kill(thispid, SIGTERM))
+		if (kill(thispid, SIGTERM))
 			perror("ERROR: dkg-tcpip-common (kill)");
 		if (opt_verbose)
 			std::cerr << "INFO: waitpid(" << thispid << ", NULL, 0)" <<
@@ -998,3 +1071,33 @@ void tcpip_done
 	}
 }
 
+int run_tcpip
+	(const size_t peers,
+	 unsigned long int stpo,
+	 const std::string &hostname,
+	 const std::string &port)
+{
+	assert(peers <= DKGPG_MAX_N);
+	int ret = 0;
+	if (port.length())
+		stpo = strtoul(port.c_str(), NULL, 10); // get start port from options
+	if ((stpo < 1) || (stpo > 65535))
+	{
+		std::cerr << "ERROR: no valid TCP start port given" << std::endl;
+		return -1;
+	}
+	tcpip_init(hostname);
+	tcpip_bindports((uint16_t)stpo, false);
+	tcpip_bindports((uint16_t)stpo, true);
+	while (tcpip_connect((uint16_t)stpo, false) < peers)
+		sleep(1);
+	while (tcpip_connect((uint16_t)stpo, true) < peers)
+		sleep(1);
+	tcpip_accept();
+	tcpip_fork();
+	ret = tcpip_io();
+	tcpip_close();
+	tcpip_done();
+	return ret;
+}
+
